@@ -460,7 +460,8 @@ my %sKeywordStruct;
     NOTES => q{
         Image region metadata defined by the MWG 2.0 specification.  These tags
         may be accessed without the need to load the MWG Composite tags above.  See
-        L<http://www.metadataworkinggroup.org/> for the official specification.
+        L<https://web.archive.org/web/20181006115950/http://www.metadataworkinggroup.org/specs/>
+        for the official specification.
     },
     Regions => {
         Name => 'RegionInfo',
@@ -488,8 +489,9 @@ my %sKeywordStruct;
         Hierarchical keywords metadata defined by the MWG 2.0 specification. 
         ExifTool unrolls keyword structures to an arbitrary depth of 6 to allow
         individual levels to be accessed with different tag names, and to avoid
-        infinite recursion.  See L<http://www.metadataworkinggroup.org/> for the
-        official specification.
+        infinite recursion.  See
+        L<https://web.archive.org/web/20181006115950/http://www.metadataworkinggroup.org/specs/>
+        for the official specification.
     },
     # arbitrarily define only the first 6 levels of the keyword hierarchy
     Keywords => {
@@ -527,7 +529,8 @@ my %sKeywordStruct;
     NAMESPACE => 'mwg-coll',
     NOTES => q{
         Collections metadata defined by the MWG 2.0 specification.  See
-        L<http://www.metadataworkinggroup.org/> for the official specification.
+        L<https://web.archive.org/web/20181006115950/http://www.metadataworkinggroup.org/specs/>
+        for the official specification.
     },
     Collections => {
         FlatName => '',
@@ -745,7 +748,7 @@ must be loaded explicitly as described above.
 
 =head1 AUTHOR
 
-Copyright 2003-2019, Phil Harvey (phil at owl.phy.queensu.ca)
+Copyright 2003-2020, Phil Harvey (philharvey66 at gmail.com)
 
 This library is free software; you can redistribute it and/or modify it
 under the same terms as Perl itself.
